# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from enum import Enum, auto


class GuildFeature(Enum):
    """Represents Guild Features strings.

    Attributes
    ----------
    ANIMATED_ICON:
        Guild can have an animated icon.
    BANNER:
        Guild can have a banner.
    COMMERCE:
        Guild can have commerce.
    COMMUNITY:
        Guild is a community.
    DISCOVERABLE:
        Guild can be discovered.
    FEATURABLE:
        Guild can be featureurable.Guild can have an invite splash.
    INVITE_SPLASH:
        Guild can be featureurable.
    MEMBER_VERIFICATION_GATE_ENABLED:
        Guild has member verification enabled.
    NEWS:
        Guild has a news channel.
    PARTNERED:
        Guild is a partner.
    PREVIEW_ENABLED:
        Guild has preview enabled.
    VANITY_URL:
        Guild has a vanity url.
    VERIFIED:
        Guild is verified.
    VIP_REGIONS:
        Guild can have VIP regions.
    WELCOME_SCREEN_ENABLED:
        Guild has welcome screen enabled.
    TICKETED_EVENTS_ENABLED:
        Guild has ticketed events enabled.
    MONETIZATION_ENABLED:
        Guild has monetization enabled.
    MORE_STICKERS:
        Guild can have more stickers.
    THREE_DAY_THREAD_ARCHIVE:
        Guild can have three day archive time for threads.
    SEVEN_DAY_THREAD_ARCHIVE:
        Guild can have seven day archive time for threads.
    PRIVATE_THREADS:
        Guild can have private threads.
    """
    ANIMATED_ICON = auto()
    BANNER = auto()
    COMMERCE = auto()
    COMMUNITY = auto()
    DISCOVERABLE = auto()
    FEATURABLE = auto()
    INVITE_SPLASH = auto()
    MEMBER_VERIFICATION_GATE_ENABLED = auto()
    NEWS = auto()
    PARTNERED = auto()
    PREVIEW_ENABLED = auto()
    VANITY_URL = auto()
    VERIFIED = auto()
    VIP_REGIONS = auto()
    WELCOME_SCREEN_ENABLED = auto()
    TICKETED_EVENTS_ENABLED = auto()
    MONETIZATION_ENABLED = auto()
    MORE_STICKERS = auto()
    THREE_DAY_THREAD_ARCHIVE = auto()
    SEVEN_DAY_THREAD_ARCHIVE = auto()
    PRIVATE_THREADS = auto()



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .api_object import APIObject, HTTPMeta
from .conversion import convert
from .event_mgr import EventMgr
from .directory import chdir
from .extraction import get_index
from .insertion import should_pass_cls, should_pass_ctx
from .signature import get_params, get_signature_and_params
from .snowflake import Snowflake
from .tasks import Task, TaskScheduler
from .timestamp import Timestamp

from .types import (
    APINullable, Coro, MISSING, Descripted, MissingType, Choices,
    choice_value_types, CheckFunction
)


__all__ = (
    "APINullable", "APIObject", "CheckFunction", "Choices", "Coro",
    "Descripted", "EventMgr", "HTTPMeta", "MISSING", "MissingType",
    "Snowflake", "Task", "TaskScheduler", "Timestamp", "chdir",
    "choice_value_types", "convert", "get_index", "get_params",
    "get_signature_and_params", "should_pass_cls", "should_pass_ctx"
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ..guild.guild import Guild
from ..guild.member import GuildMember
from ..user import User
from ...utils.api_object import APIObject
from ...utils.conversion import construct_client_dict
from ...utils.types import MISSING, APINullable

if TYPE_CHECKING:
    from typing import Any, List, Optional

    from ..guild.role import Role
    from ..message.emoji import Emoji
    from ..message.sticker import Sticker
    from ...utils.snowflake import Snowflake
    from ...utils.timestamp import Timestamp
    from .presence import PresenceUpdateEvent


@dataclass
class GuildBanAddEvent(APIObject):
    """Sent when a user is banned from a guild.

    Attributes
    ----------
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the guild
    user: :class:`~pincer.objects.user.user.User`
        The banned user
    """

    guild_id: Snowflake
    user: User


@dataclass
class GuildBanRemoveEvent(APIObject):
    """Sent when a user is unbanned from a guild.

    Attributes
    ----------
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the guild
    user: :class:`~pincer.objects.user.user.User`
        The unbanned user
    """

    guild_id: Snowflake
    user: User


@dataclass
class GuildEmojisUpdateEvent(APIObject):
    """Sent when a guild's emojis have been updated.

    Attributes
    ----------
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the guild
    emojis: List[:class:`~pincer.objects.message.emoji.Emoji`]
        Array of emojis
    """

    guild_id: Snowflake
    emojis: List[Emoji]


@dataclass
class GuildStickersUpdateEvent(APIObject):
    """Sent when a guild's stickers have been updated.

    Attributes
    ----------
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the guild
    stickers: List[:class:`~pincer.objects.message.sticker.Sticker`]
        Array of stickers
    """

    guild_id: Snowflake
    stickers: List[Sticker]


@dataclass
class GuildIntegrationsUpdateEvent(APIObject):
    """Sent when a guild integration is updated.

    Attributes
    ----------
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the guild whose integrations were updated
    """

    guild_id: Snowflake


@dataclass
class GuildMemberAddEvent(GuildMember):
    """
    Sent when a user joins a guild.

    Attributes
    ----------
    guild_id: :class:`Snowflake`
        ID of the guild that the user joined.
    """

    # NOTE: This isn't a default value. I set it to this because you can't
    # have fields without default values after regular fields. Apparently that
    # carries over when you inherit from a dataclass.
    guild_id: Snowflake = 0


@dataclass
class GuildMemberRemoveEvent(APIObject):
    """Sent when a user is removed from a guild (leave/kick/ban).

    Attributes
    ----------
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        the id of the guild
    user: :class:`~pincer.objects.user.user.User`
        the user who was removed
    """

    guild_id: Snowflake
    user: User

    def __post_init__(self):
        self.user = User.from_dict(
            construct_client_dict(self._client, self.user)
        )


@dataclass
class GuildMemberUpdateEvent(APIObject):
    """Sent when a guild member is updated.
    This will also fire when the user object
    of a guild member changes.

    Attributes
    ----------
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        the id of the guild
    roles: List[:class:`~pincer.utils.snowflake.Snowflake`]
        user role ids
    user: :class:`~pincer.objects.user.user.User`
        the user
    nick: APINullable[Optional[:class:`str`]]
        nickname of the user in the guild
    joined_at: Optional[:class:`~pincer.utils.timestamp.Timestamp`]
        when the user joined the guild
    premium_since: APINullable[Optional[:class:`~pincer.utils.timestamp.Timestamp`]]
        when the user started boosting the guild
    deaf: APINullable[:class:`bool`]
        whether the user is deafened in voice channels
    mute: APINullable[:class:`bool`]
        whether the user is muted in voice channels
    pending: APINullable[:class:`bool`]
        whether the user has not yet passed the guild's
        Membership Screening requirements
    """
    # noqa: E501

    guild_id: Snowflake
    roles: List[Snowflake]
    user: User
    nick: APINullable[Optional[str]] = MISSING
    joined_at: Optional[Timestamp] = None
    premium_since: APINullable[Optional[Timestamp]] = MISSING
    deaf: APINullable[bool] = MISSING
    mute: APINullable[bool] = MISSING
    pending: APINullable[bool] = MISSING

    def __post_init__(self):
        self.user = User.from_dict(
            construct_client_dict(self._client, self.user)
        )


@dataclass
class GuildMembersChunkEvent(APIObject):
    """Sent in response to Guild Request Members.
    You can use the ``chunk_index`` and ``chunk_count``
    to calculate how many chunks are left for your request.

    Attributes
    ----------
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the guild
    members: List[:class:`~pincer.objects.guild.member.GuildMember`]
        Set of guild members
    chunk_index: :class:`int`
        The chunk index in the expected chunks for this response
        (0 <= chunk_index < chunk_count)
    chunk_count: :class:`int`
        The total number of expected chunks for this response
    not_found: APINullable[List[Any]]
        If passing an invalid id to ``REQUEST_GUILD_MEMBERS``,
        it will be returned here
    presences: APINullable[:class:`~pincer.objects.events.presence.PresenceUpdateEvent`]
        If passing true to ``REQUEST_GUILD_MEMBERS``, presences
        of the returned members will be here
    nonce: APINullable[:class:`str`]
        The nonce used in the Guild Members Request
    """
    # noqa: E501
    guild_id: Snowflake
    members: List[GuildMember]
    chunk_index: int
    chunk_count: int

    not_found: APINullable[List[Any]] = MISSING
    presences: APINullable[PresenceUpdateEvent] = MISSING
    nonce: APINullable[str] = MISSING

    def __post_init__(self):
        self.members = [
            GuildMember.from_dict(construct_client_dict(self._client, member))
            for member in self.members
        ]


@dataclass
class GuildRoleCreateEvent(APIObject):
    """Sent when a guild role is created.

    Attributes
    ----------
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the guild
    role: :class:`~pincer.objects.guild.role.Role`
        The role created
    """

    guild_id: Snowflake
    role: Role


@dataclass
class GuildRoleUpdateEvent(APIObject):
    """Sent when a guild role is updated.

    Attributes
    ----------
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the guild
    role: :class:`~pincer.objects.guild.role.Role`
        The role updated
    """

    guild_id: Snowflake
    role: Role


@dataclass
class GuildRoleDeleteEvent(APIObject):
    """Sent when a guild role is deleted.

    Attributes
    ----------
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the guild
    role_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the role
    """

    guild_id: Snowflake
    role_id: Snowflake


@dataclass
class GuildStatusEvent(APIObject):
    """
    Sent when a subscribed server's state changes

    Attributes
    ----------
    guild : :class:`Guild`
        guild with requested id

    online : :class:`int`
        number of online users in guild (deprecated; always 0)
    """

    guild: Guild
    online: int = 0



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a user in a subscribed voice channel stops speaking"""

from ..core.dispatch import GatewayDispatch
from ..objects.events.voice import SpeakingStopEvent
from ..utils.conversion import construct_client_dict
from ..utils.types import Coro


async def speaking_stop_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_speaking_stop`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the speaking stop event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`Snowflake`]]
        ``on_speaking_stop`` and a ``Snowflake`` (user_id)
    """
    return "on_speaking_stop", [
        SpeakingStopEvent.from_dict(construct_client_dict(self, payload.data))
    ]


def export() -> Coro:
    return speaking_stop_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject

if TYPE_CHECKING:
    from ...utils.snowflake import Snowflake


@dataclass
class WebhooksUpdateEvent(APIObject):
    """
    Sent when a guild's channel webhook
    is created, updated, or deleted.

    Attributes
    ----------
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        id of the guild
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        id of the channel
    """
    guild_id: Snowflake
    channel_id: Snowflake



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, TYPE_CHECKING

from ...utils import APIObject

if TYPE_CHECKING:
    from ..user.user import User


@dataclass
class Ban(APIObject):
    """Representation of the Discord Ban object

    Attributes
    ----------
    reason: Optional[:class:`str`]
        The reason for the ban
    user: :class:`~pincer.objects.user.user.User`
        The banned user
    """
    reason: Optional[str]
    user: User



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when an integration is created"""

from ..core.dispatch import GatewayDispatch
from ..objects.events.integration import IntegrationCreateEvent
from ..utils.conversion import construct_client_dict
from ..utils.types import Coro


async def integration_create_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_integration_create``,

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the integration create event

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.events.integration.IntegrationCreateEvent`]]
        ``on_integration_create`` and an ``IntegrationCreateEvent``
    """
    return "on_integration_create", [
        IntegrationCreateEvent.from_dict(construct_client_dict(self, payload.data))
    ]


def export() -> Coro:
    return integration_create_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a user started typing in a channel"""

from ..core.dispatch import GatewayDispatch
from ..objects.events.typing_start import TypingStartEvent
from ..utils.conversion import construct_client_dict
from ..utils.types import Coro


async def typing_start_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_typing_start`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the typing start event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.events.typing_start.TypingStartEvent`]]
        ``on_typing_start`` and a ``TypingStartEvent``
    """
    return "on_typing_start", [
        TypingStartEvent.from_dict(construct_client_dict(self, payload.data))
    ]


def export() -> Coro:
    return typing_start_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import APINullable, MISSING

if TYPE_CHECKING:
    from typing import List

    from ..message.emoji import Emoji
    from ..guild.member import GuildMember
    from ...utils.snowflake import Snowflake


@dataclass
class MessageDeleteEvent(APIObject):
    """Sent when a message is deleted.

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the message
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the channel
    guild_id: APIObject[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the guild
    """
    id: Snowflake
    channel_id: Snowflake

    guild_id: APINullable[Snowflake] = MISSING


@dataclass
class MessageDeleteBulkEvent(APIObject):
    """Sent when multiple messages are deleted at once.

    Attributes
    ----------
    ids: List[:class:`~pincer.utils.snowflake.Snowflake`]
        The ids of the messages
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the channel
    guild_id: APIObject[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the guild
    """
    ids: List[Snowflake]
    channel_id: Snowflake

    guild_id: APINullable[Snowflake] = MISSING


@dataclass
class MessageReactionAddEvent(APIObject):
    """Sent when a user adds a reaction to a message.

    Attributes
    ----------
    user_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the user
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the channel
    message_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the message
    emoji: :class:`~pincer.objects.message.emoji.Emoji`
        The emoji used to react
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the guild
    member: :class:`~pincer.objects.guild.member.GuildMember`
        The member who reacted if this happened in a guild
    """
    user_id: Snowflake
    channel_id: Snowflake
    message_id: Snowflake
    emoji: Emoji

    guild_id: APINullable[Snowflake] = MISSING
    member: APINullable[GuildMember] = MISSING


@dataclass
class MessageReactionRemoveEvent(APIObject):
    """Sent when a user removes a reaction from a message.

    Attributes
    ----------
    user_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the user
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the channel
    message_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the message
    emoji: :class:`~pincer.objects.message.emoji.Emoji`
        The emoji used to react
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the guild
    """
    user_id: Snowflake
    channel_id: Snowflake
    message_id: Snowflake
    emoji: Emoji

    guild_id: APINullable[Snowflake] = MISSING


@dataclass
class MessageReactionRemoveAllEvent(APIObject):
    """Sent when a user explicitly removes all reactions from a message.

    Attributes
    ----------
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the channel
    message_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the message
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the guild
    """
    channel_id: Snowflake
    message_id: Snowflake
    guild_id: APINullable[Snowflake] = MISSING


@dataclass
class MessageReactionRemoveEmojiEvent(APIObject):
    """Sent when a bot removes all instances of a given
    emoji from the reactions of a message.

    Attributes
    ----------
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the channel
    message_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the message
    emoji: :class:`~pincer.objects.message.emoji.Emoji`
        The emoji that was removed
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the guild
    """
    channel_id: Snowflake
    message_id: Snowflake
    emoji: Emoji

    guild_id: APINullable[Snowflake] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from typing import Optional


@dataclass
class GatewayConfig:
    """This file is to make maintaining the library its gateway
    configuration easier.
    """
    socket_base_url: str = "wss://gateway.discord.gg/"
    version: int = 9
    encoding: str = "json"
    compression: Optional[str] = "zlib-stream"

    @classmethod
    def uri(cls) -> str:
        """
        Returns
        -------
        :class:`str`:
            The GatewayConfig's uri.
        """
        return (
            f"{cls.socket_base_url}"
            f"?v={cls.version}"
            f"&encoding={cls.encoding}"
        ) + f"&compress={cls.compression}" * cls.compressed()

    @classmethod
    def compressed(cls) -> bool:
        """
        Returns
        -------
        :class:`bool`:
            Whether the Gateway should compress payloads or not.
        """
        return cls.compression in ["zlib-stream", "zlib-payload"]



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from ...utils.types import APINullable
    from ..message.emoji import Emoji


class ButtonStyle(IntEnum):
    """Buttons come in a variety of styles to convey different types of actions.
    These styles also define what fields are valid for a button.

    Attributes
    ----------
    Primary:
        - color: blurple
        - required_field: custom_id
    Secondary:
        - color: gray
        - required_field: custom_id
    Success:
        - color: green
        - required_field: custom_id
    Danger:
        - color: red
        - required_field: custom_id
    Link:
        - color: gray, navigates to a URL
        - required_field: url
    """
    PRIMARY = 1
    SECONDARY = 2
    SUCCESS = 3
    DANGER = 4
    LINK = 5


@dataclass
class Button(APIObject):
    """Represents a Discord Button object.
    Buttons are interactive components that render on messages.

    They can be clicked by users,
    and send an interaction to your app when clicked.

    Attributes
    ----------
    type: :class:`int`
        `2` for a button
    style: :class:`~pincer.objects.message.button.ButtonStyle`
        one of button styles
    label: APINullable[:class:`str`]
        text that appears on the button, max 80 characters
    emoji: APINullable[:class:`~pincer.objects.message.emoji.Emoji`]
        ``name``, ``id``, and ``animated``
    custom_id: APINullable[:class:`str`]
        A developer-defined identifier for the button,
        max 100 characters
    url: APINullable[:class:`str`]
        A url for link-style buttons
    disabled: APINullable[:class:`bool`]
        Whether the button is disabled (default `False`)
    """
    type: int
    style: ButtonStyle

    label: APINullable[str] = MISSING
    emoji: APINullable[Emoji] = MISSING
    custom_id: APINullable[str] = MISSING
    url: APINullable[str] = MISSING
    disabled: APINullable[bool] = False



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING, APINullable

if TYPE_CHECKING:
    from ...utils.snowflake import Snowflake
    from ...utils.timestamp import Timestamp


@dataclass
class ChannelPinsUpdateEvent(APIObject):
    """Sent when a message is pinned or unpinned in a text channel.
    This is not sent when a pinned message is deleted.

    Attributes
    ----------
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the channel
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the guild
    last_pin_timestamp: APINullable[:class:`~pincer.utils.timestamp.Timestamp`]
        The time at which the most recent pinned message was pinned
    """
    channel_id: Snowflake

    guild_id: APINullable[Snowflake] = MISSING
    last_pin_timestamp: APINullable[Timestamp] = MISSING




# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import List, Optional

    from .guild import Guild
    from ..app.application import Application
    from ..guild.channel import Channel
    from ..guild.member import GuildMember
    from ..user.user import User
    from ...utils.timestamp import Timestamp
    from ...utils.types import APINullable


class InviteTargetType(IntEnum):
    """Represents the type of the invite.

    Attributes
    ----------
    STREAM:
        A normal Discord invite, eg for a channel or guild.
    EMBEDDED_APPLICATION:
        An embedded application invite, eg poker-night etc.
    """
    STREAM = 1
    EMBEDDED_APPLICATION = 2


@dataclass
class InviteStageInstance(APIObject):
    """Represents an invite for a Discord stages channel.

    Attributes
    ----------
    members: List[:class:`~pincer.objects.guild.member.GuildMember`]
        the members speaking in the Stage
    participant_count: :class:`int`
        the number of users in the Stage
    speaker_count: :class:`int`
        the number of users speaking in the Stage
    topic: :class:`str`
        the topic of the Stage instance (1-120 characters)
    """
    members: List[GuildMember]
    participant_count: int
    speaker_count: int
    topic: str


@dataclass
class InviteMetadata(APIObject):
    """Extra information about an invite, will extend the invite object.

    Attributes
    ----------
    uses: :class:`int`
        number of times this invite has been used
    max_uses: :class:`int`
        Max number of times this invite can be used
    max_age: :class:`int`
        Duration (in seconds) after which the invite expires
    temporary:  :class:`bool`
        Whether this invite only grants temporary membership
    created_at: :class:`~pincer.utils.timestamp.Timestamp`
        When this invite was created
    """
    uses: int
    max_uses: int
    max_age: int
    temporary: bool
    created_at: Timestamp


@dataclass
class Invite(APIObject):
    """Represents a Discord invite.

    Attributes
    ----------
    channel: :class:`~pincer.objects.guild.channel.Channel`
        The channel this invite is for
    code: :class:`str`
        The invite code (unique ID)
    approximate_member_count: APINullable[:class:`int`]
        Approximate count of total members, returned from the GET
        /invites/<code> endpoint when with_counts is true
    approximate_presence_count: APINullable[:class:`int`]
        Approximate count of online members, returned from the GET
        /invites/<code> endpoint when with_counts is true
    expires_at: APINullable[Optional[:class:`~pincer.utils.timestamp.Timestamp`]]
        The expiration date of this invite, returned from the GET
        /invites/<code> endpoint when with_expiration is true
    inviter: APINullable[:class:`~pincer.objects.user.user.User`]
        The user who created the invite
    guild: :class:`~pincer.objects.guild.guild.Guild`
        The guild this invite is for
    stage_instance: :class:`~pincer.objects.guild.invite.InviteStageInstance`
        Stage instance data if there is a public Stage instance in the Stage
        channel this invite is for
    target_type: :class:`~pincer.objects.guild.invite.InviteTargetType`
        The type of target for this voice channel invite
    target_user: :class:`~pincer.objects.user.user.User`
        The user whose stream to display for this voice channel stream invite
    target_application: :class:`~pincer.objects.app.application.Application`
        The embedded application to open for this voice channel embedded
        application invite
    """
    # noqa: E501

    channel: Channel
    code: str

    approximate_member_count: APINullable[int] = MISSING
    approximate_presence_count: APINullable[int] = MISSING
    expires_at: APINullable[Optional[Timestamp]] = MISSING
    inviter: APINullable[User] = MISSING
    guild: APINullable[Guild] = MISSING
    stage_instance: APINullable[InviteStageInstance] = MISSING
    target_type: APINullable[InviteTargetType] = MISSING
    target_user: APINullable[User] = MISSING
    target_application: APINullable[Application] = MISSING

    def __str__(self) -> str:
        return self.link

    @property
    def link(self):
        return f"https://discord.gg/{self.code}"



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from json import dumps
from typing import TYPE_CHECKING

from aiohttp import FormData

from ..message.file import File
from ...exceptions import CommandReturnIsEmpty

if TYPE_CHECKING:
    from typing import Dict, Union, List, Optional, Tuple

    from aiohttp import Payload

    from .embed import Embed
    from .component import MessageComponent
    from ..app.interactions import InteractionFlags
    from ..message.user_message import AllowedMentions
    from ...objects.app import CallbackType

PILLOW_IMPORT = True

try:
    from PIL.Image import Image
except (ModuleNotFoundError, ImportError):
    PILLOW_IMPORT = False


@dataclass
class Message:
    """A discord message that will be send to discord

    Attributes
    ----------
    content: :class:`str`
        The text in the message.
        |default| ``""``
    attachments: Optional[List[:class:`~pincer.objects.message.file.File`]]
        Attachments on the message. This is a File object. You can also attach
        a Pillow Image or string. Pillow images will be converted to PNGs. They
        will use the naming scheme ``image%`` where % is the images index in the
        attachments array. Strings will be read as a filepath. The name of the
        file that the string points to will be used as the name.
    tts: Optional[:class:`bool`]
        Whether the message should be spoken to the user.
        |default| :data:`False`
    embeds: Optional[List[:class:`~pincer.objects.message.embed.Embed`]]
        Embed attached to the message. This is an Embed object.
    allowed_mentions: Optional[:class:`~pincer.objects.message.message.AllowedMentions`]
        The allowed mentions for the message.
    components: Optional[List[:class:`~pincer.objects.message.component.MessageComponent`]]
        The components of the message.
    flags: Optional[:class:`~pincer.objects.app.interactions.InteractionFlags`]
        The interaction flags for the message.
    type: Optional[:class:`~pincer.objects.app.interaction_base.CallbackType`]
        The type of the callback.
    """
    # noqa: E501

    content: str = ''
    attachments: Optional[List[File]] = None
    tts: Optional[bool] = False
    embeds: Optional[List[Embed]] = None
    allowed_mentions: Optional[AllowedMentions] = None
    components: Optional[List[MessageComponent]] = None
    flags: Optional[InteractionFlags] = None
    delete_after: Optional[float] = None

    def __post_init__(self):
        if self.delete_after and self.delete_after < 0:
            raise ValueError(
                "Message can not be deleted after a negative amount of "
                "seconds!"
            )

        if not self.attachments:
            return

        attachment = []

        for count, value in enumerate(self.attachments):
            if isinstance(value, File):
                attachment.append(value)
            elif PILLOW_IMPORT and isinstance(value, Image):
                attachment.append(File.from_pillow_image(
                    value,
                    f"image{count}.png",
                ))
            elif isinstance(value, str):
                attachment.append(File.from_file(value))
            else:
                raise ValueError(f"Attachment {count} is invalid type.")

        self.attachments = attachment

    @property
    def isempty(self) -> bool:
        """:class:`bool`: If the message is empty."""

        return (
            len(self.content) < 1
            and not self.embeds
            and not self.attachments
        )

    def to_dict(self):

        allowed_mentions = (
            self.allowed_mentions.to_dict()
            if self.allowed_mentions else {}
        )

        # Attachments aren't serialized
        # because they are not sent as part of the json
        resp = {
            "content": self.content,
            "tts": self.tts,
            "flags": self.flags,
            "embeds": [embed.to_dict() for embed in (self.embeds or [])],
            "allowed_mentions": allowed_mentions,
            "components": [
                components.to_dict() for components in (self.components or [])
            ]
        }

        # IDE does not recognise return type of filter properly.
        # noinspection PyTypeChecker
        return dict(filter(
            lambda kv: kv[1],
            resp.items()
        ))

    def serialize(
        self, message_type: Optional[CallbackType] = None
    ) -> Tuple[str, Union[Payload, Dict]]:
        """
        Parameters
        ----------
        message_type : Optional[:class:`pincer.objects.app.CallbackType`]
            Callback type of message.

        Returns
        -------
        Tuple[str, Union[Payload, Dict]]
            (content_type, data)

        Raises
        ------
        :class:`pincer.exceptions.CommandReturnIsEmpty`
            Command does not have content, an embed, or attachment.
        """
        if self.isempty:
            raise CommandReturnIsEmpty("Cannot return empty message.")

        json_payload = self.to_dict()

        if message_type is not None:
            json_data = json_payload
            json_payload = {
                "data": json_data,
                "type": message_type
            }

        if not self.attachments:
            return "application/json", json_payload

        form = FormData()
        form.add_field("payload_json", dumps(json_payload))

        for file in self.attachments:
            form.add_field("file", file.content, filename=file.filename)

        payload = form()
        return payload.headers["Content-Type"], payload



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when anyone is added to or removed from a thread"""
from typing import List

from ..core.dispatch import GatewayDispatch
from ..objects import ThreadMember
from ..objects.events.thread import ThreadMembersUpdateEvent
from ..utils import Timestamp
from ..utils.conversion import construct_client_dict


async def thread_members_update_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_thread_members_update`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the thread members update event.


    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.events.thread.ThreadMembersUpdateEvent`]]
        ``on_thread_members_update`` and an ``ThreadMembersUpdateEvent``
    """

    added_members: List[ThreadMember] = [
        ThreadMember.from_dict(construct_client_dict(
            self,
            {
                "join_timestamp": Timestamp(added_member.pop("join_timestamp")),
                **added_member
            }
        ))
        for added_member in payload.data.pop("added_members")
    ]

    return "on_thread_members_update", [
        ThreadMembersUpdateEvent.from_dict(
            {
                "added_members": added_members,
                **payload.data
            }
        )
    ]


def export():
    return thread_members_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.
from typing import Optional


class PincerError(Exception):
    """Base exception class for all Pincer errors"""


class InvalidPayload(PincerError):
    """Exception which gets thrown if an invalid payload has been received.
    This means that the data of the payload did not match the expected
    format and/or didn't contain the the expected values.
    """


class UnhandledException(PincerError):
    """Exception which gets thrown if an exception wasn't handled.

    Please create an issue on our github
    if this exception gets thrown.
    """

    def __init__(self, specific: str):
        super(UnhandledException, self).__init__(
            specific + " Please report this to the library devs."
        )


class NoExportMethod(PincerError):
    """Exception which gets raised when an `export` method is expected but
    not found in a module.
    """


class CogError(PincerError):
    """Exception base class for errors related to Cogs."""


class CogNotFound(CogError):
    """Exception which gets raised when a cog is trying to be
    loaded/unloaded but is nonexistent.
    """


class CogAlreadyExists(CogError):
    """Exception which gets raised when a cog is already loaded, but is
    trying to be reloaded!
    """


class NoValidSetupMethod(CogError):
    """Exception which gets raised when an `setup` function is expected but
    none was found!
    """


class TooManySetupArguments(CogError):
    """Exception which gets raised when too many arguments were requested
    in a cog its setup function.
    """


class NoCogManagerReturnFound(CogError):
    """Exception which gets raised when no cog return was found from the
    setup function. (are you missing a return statement?)
    """


class CommandError(PincerError):
    """Base class for exceptions which are related to commands."""


class CommandCooldownError(CommandError):
    """Exception which gets raised when a command cooldown has not been
    breached.

    Attributes
    ----------
    ctx: :class:`~objects.message.context.MessageContext`
        The context of the error
    """

    def __init__(self, message: str, context):
        self.ctx = context
        super(CommandCooldownError, self).__init__(message)


class CommandIsNotCoroutine(CommandError):
    """Exception raised when the provided command call is not a coroutine."""


class CommandAlreadyRegistered(CommandError):
    """The command which you are trying to register is already registered."""


class CommandDescriptionTooLong(CommandError):
    """The provided command description is too long,
    as it exceeds 100 characters.
    """


class TooManyArguments(CommandError):
    """A command can have a maximum of 25 arguments.
    If this number of arguments gets exceeded, this exception will be raised.
    """


class InvalidArgumentAnnotation(CommandError):
    """The provided argument annotation is not known, so it cannot be used."""


class CommandReturnIsEmpty(CommandError):
    """Cannot return an empty string to an interaction."""


class InvalidCommandGuild(CommandError):
    """The provided guild id not not valid."""


class InteractionDoesNotExist(CommandError):
    """The action which you are trying to perform requires an
    interaction to be created/sent to discord. But this has not been
    done yet!
    """


class UseFollowup(CommandError):
    """A reply has already been sent, please use a followup to
    continue replying.
    """


class InteractionAlreadyAcknowledged(CommandError):
    """The command has already been acknowledged by discord.
    This can be because a reply or ack has already been sent!
    """


class InteractionTimedOut(CommandError):
    """Discord had to wait too long for a response from your command!
    The discord wait time can be extended by using the
    :func:`~pincer.objects.app.interaction.Interaction.ack` function in
    the :attr:`~pincer.objects.message.context.MessageContext.interaction`
    property.
    """


class InvalidCommandName(CommandError):
    """Exception raised when the command is considered invalid.
    This is caused by a name that doesn't match the command name regex.
    """


class InvalidEventName(PincerError):
    """Exception raised when the event name is not a valid event.
    This can be because the event name did not begin with an ``on_`` or
    because its not a valid event in the library.
    """


class InvalidUrlError(PincerError, ValueError):
    """Exception raised when an invalid url has been provided."""


class EmbedFieldError(PincerError, ValueError):
    """Exception that is raised when an embed field is too large."""

    @classmethod
    def from_desc(cls, _type: str, max_size: int, cur_size: int):
        """Create an instance by description.

        Parameters
        ----------
        _type :class:`str`
            The type/name of the field.
        max_si :class:`int`
            The maximum size of the field.
        cur_size :class:`int`
            The current size of the field.
        """
        return cls(
            f"{_type} can have a maximum length of {max_size}."
            f" (Current size: {cur_size})"
        )


class TaskError(PincerError):
    """Base class for exceptions that are related to tasks.

    Attributes
    ----------
    task: :class:`~utils.tasks.Task`
        The task that raised the exception.
    """

    def __init__(self, message: str, task=None):
        self.task = task
        super().__init__(message)


class TaskAlreadyRunning(TaskError):
    """Exception that is raised when the user tries to start a running task."""


class TaskCancelError(TaskError):
    """Exception that is raised when a task cannot be cancelled."""


class TaskIsNotCoroutine(TaskError):
    """Exception that is raised when the provided function for a task is not
    a coroutine.
    """


class TaskInvalidDelay(TaskError):
    """Exception that is raised when the provided delay is invalid."""


class DispatchError(PincerError):
    """Base exception class for all errors which are specifically related
    to the dispatcher.
    """


class _InternalPerformReconnectError(DispatchError):
    """Internal helper exception which on raise lets the client reconnect."""


class DisallowedIntentsError(DispatchError):
    """Invalid gateway intent got provided.
    Make sure your client has the enabled intent.
    """


class InvalidTokenError(DispatchError, ValueError):
    """Exception raised when the authorization token is invalid."""

    def __init__(self, hint: Optional[str] = None):
        hint = hint or ''

        super(InvalidTokenError, self).__init__(
            "The given token is not a valid token.\n" + hint
        )


class HeartbeatError(DispatchError):
    """Exception raised due to a problem with websocket heartbeat."""


class UnavailableGuildError(PincerError):
    """Exception raised due to a guild being unavailable.
    This is caused by a discord outage.
    """


class TimeoutError(PincerError):
    """Exception raised when :class:`~pincer.utils.event_mgr.EventMgr`
    `wait_for` and `loop_for` methods time out
    """


# Discord HTTP Errors
# `developers/docs/topics/opcodes-and-status-codes#http`


class HTTPError(PincerError):
    """HTTP Exception base class."""


class NotModifiedError(HTTPError):
    """Error code 304."""


class BadRequestError(HTTPError):
    """Error code 400."""


class UnauthorizedError(HTTPError):
    """Error code 401."""


class ForbiddenError(HTTPError):
    """Error code 403."""


class NotFoundError(HTTPError):
    """Error code 404."""


class MethodNotAllowedError(HTTPError):
    """Error code 405."""


class RateLimitError(HTTPError):
    """Error code 429."""


class GatewayError(HTTPError):
    """Error code 502."""


class ServerError(HTTPError):
    """Error code 5xx."""



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a message is deleted in a subscribed text channel"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.events.message import MessageDeleteEvent
from ..utils.conversion import construct_client_dict

if TYPE_CHECKING:
    from typing import List, Tuple

    from ..core.dispatch import GatewayDispatch


async def on_message_delete_middleware(
    self,
    payload: GatewayDispatch
) -> Tuple[str, List[MessageDeleteEvent]]:
    """|coro|
    Middleware for ``on_message_delete`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the message delete event

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.events.message.MessageDeleteEvent`]
        ``on_message_delete`` and a ``MessageDeleteEvent``
    """

    return "on_message_delete", [
        MessageDeleteEvent.from_dict(construct_client_dict(self, payload.data))
    ]


def export():
    return on_message_delete_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from ...utils.api_object import APIObject


@dataclass
class SessionStartLimit(APIObject):
    """Represents a Discord Session Start Limit object

    Attributes
    ----------
    total: :class:`int`
        The total number of session starts the current user is allowed
    remaining: :class:`int`
        The remaining number of session starts
        the current user is allowed
    reset_after: :class:`int`
        The number of milliseconds after which the limit resets
    max_concurrency: :class:`int`
        The number of identify requests allowed per 5 seconds
    """
    total: int
    remaining: int
    reset_after: int
    max_concurrency: int



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from ...utils.types import APINullable
    from ...utils.snowflake import Snowflake


@dataclass
class MessageReference(APIObject):
    """Represents a Discord Message Reference object

    Attributes
    ----------
    message_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Id of the originating message
    channel_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Id of the originating message's channel
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Id of the originating message's guild
    fail_if_not_exists: APINullable[:class:`bool`]
        When sending, whether to error if the referenced message doesn't
        exist instead of sending as a normal (non-reply) message,
        default true
    """
    message_id: APINullable[Snowflake] = MISSING
    channel_id: APINullable[Snowflake] = MISSING
    guild_id: APINullable[Snowflake] = MISSING
    fail_if_not_exists: APINullable[bool] = True



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING, APINullable

if TYPE_CHECKING:
    from typing import List, Optional, Tuple
    from ..user.user import User
    from ...utils.snowflake import Snowflake


class ActivityType(IntEnum):
    """Represents the enum of the type of activity.

    Attributes
    ----------
    GAME:
        Playing {name}; e.g. "Playing Rocket League"
    STREAMING:
        Streaming {details}; e.g. "Streaming Rocket League"; Only supports Twitch and YouTube.
    LISTENING:
        Listening to {name}; e.g. "Listening to Spotify"
    WATCHING:
        Watching {name}; e.g. "Watching YouTube Together"
    CUSTOM:
        \\{emoji} {name}; e.g. "\\:smiley: I am cool"; Not for bots; discord limitation
    COMPETING:
        Competing in {name}; e.g. "Competing in Arena World Champions"
    """
    # noqa: E501
    GAME = 0
    STREAMING = 1
    LISTENING = 2
    WATCHING = 3
    CUSTOM = 4
    COMPETING = 5


@dataclass
class ActivityTimestamp(APIObject):
    """Represents the timestamp of an activity.

    Attributes
    ----------
    start: APINullable[:class:`int`]
        Unix time (in milliseconds) of when the activity started
    end: APINullable[:class:`int`]
        Unix time (in milliseconds) of when the activity ends
    """
    start: APINullable[int] = MISSING
    end: APINullable[int] = MISSING


@dataclass
class ActivityEmoji(APIObject):
    """Represents an emoji in an activity.

    Attributes
    ----------
    name: :class:`str`
        The name of the emoji
    id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the emoji
    animated: APINullable[:class:`bool`]
        Whether this emoji is animated
    """
    name: str
    id: APINullable[Snowflake] = MISSING
    animated: APINullable[bool] = MISSING


@dataclass
class ActivityParty(APIObject):
    """Represents a party in an activity.

    Attributes
    ----------
    id: APINullable[:class:`str`]
        The id of the party
    size: APINullable[Tuple[:class:`int`, :class:`int`]]
        Array of two integers (current_size, max_size)
    """
    id: APINullable[str] = MISSING
    size: APINullable[Tuple[int, int]] = MISSING


@dataclass
class ActivityAssets(APIObject):
    """Represents an asset of an activity.

    Attributes
    ----------
    large_image: APINullable[:class:`str`]
        the id for a large asset of the activity, usually a snowflake
    large_text: APINullable[:class:`str`]
        text displayed when hovering over
        the large image of the activity
    small_image: APINullable[:class:`str`]
        the id for a small asset of the activity, usually a snowflake
    small_text: APINullable[:class:`str`]
        text displayed when hovering over
        the small image of the activity
    """
    large_image: APINullable[str] = MISSING
    large_text: APINullable[str] = MISSING
    small_image: APINullable[str] = MISSING
    small_text: APINullable[str] = MISSING


@dataclass
class ActivitySecrets(APIObject):
    """Represents a secret of an activity.

    Attributes
    ----------
    join: APINullable[:class:`str`]
        The secret for joining a party
    spectate: APINullable[:class:`str`]
        The secret for spectating a game
    match: APINullable[:class:`str`]
        The secret for a specific instanced match
    """
    join: APINullable[str] = MISSING
    spectate: APINullable[str] = MISSING
    match_: APINullable[str] = MISSING


class ActivityFlags(IntEnum):
    INSTANCE = 1 << 0
    JOIN = 1 << 1
    SPECTATE = 1 << 2
    JOIN_REQUEST = 1 << 3
    SYNC = 1 << 4
    PLAY = 1 << 5


@dataclass
class ActivityButton(APIObject):
    """When received over the gateway, the buttons field is an array
    of strings, which are the button labels. Bots cannot access
    a user's activity button URLs. When sending, the buttons field
    must be an array of this object.

    Attributes
    ----------
    label: :class:`str`
        The text shown on the button (1-32 characters)
    url: :class:`str`
        The url opened when clicking the button (1-512 characters)
    """
    label: str
    url: str


@dataclass
class Activity(APIObject):
    """Bots are only able to send ``name``, ``type``, and optionally ``url``.

    Attributes
    ----------
    name: :class:`str`
        The activity's name
    type: :class:`~pincer.objects.events.presence.ActivityType`
        Activity type
    created_at: :class:`int`
        Unix timestamp (in milliseconds) of when
        the activity was added to the user's session
    url: APINullable[Optional[:class:`str`]]
        Stream url, is validated when type is 1
    timestamps: APINullable[:class:`~pincer.objects.events.presence.ActivityTimestamp`]
        Unix timestamps for start and/or end of the game
    application_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Application id for the game
    details: APINullable[Optional[:class:`str`]]
        What the player is currently doing
    state: APINullable[Optional[:class:`str`]]
        The user's current party status
    emoji: APINullable[Optional[:class:`~pincer.objects.events.presence.ActivityEmoji`]]
        The emoji used for a custom status
    party: APINullable[:class:`~pincer.objects.events.presence.ActivityParty`]
        Information for the current party of the player
    assets: APINullable[:class:`~pincer.objects.events.presence.ActivityAssets`]
        Images for the presence and their hover texts
    secrets: APINullable[:class:`~pincer.objects.events.presence.ActivitySecrets`]
        Secrets for Rich Presence joining and spectating
    instance: APINullable[:class:`bool`]
        "nether or not the activity is an instanced game session
    flags: APINullable[:class:`~pincer.objects.events.presence.ActivityFlags`]
        Activity flags ``OR``\\d together,
        describes what the payload includes
    buttons: APINullable[List[:class:`~pincer.objects.events.presence.ActivityButton`]]
        The url button on an activity.
    """
    # noqa: E501
    name: str
    type: ActivityType
    created_at: int

    url: APINullable[Optional[str]] = MISSING
    timestamps: APINullable[ActivityTimestamp] = MISSING
    application_id: APINullable[Snowflake] = MISSING
    details: APINullable[Optional[str]] = MISSING
    state: APINullable[Optional[str]] = MISSING
    emoji: APINullable[Optional[ActivityEmoji]] = MISSING
    party: APINullable[ActivityParty] = MISSING
    assets: APINullable[ActivityAssets] = MISSING
    secrets: APINullable[ActivitySecrets] = MISSING
    instance: APINullable[bool] = MISSING
    flags: APINullable[ActivityFlags] = MISSING
    buttons: APINullable[List[ActivityButton]] = MISSING


@dataclass
class ClientStatus(APIObject):
    """Active sessions are indicated with an "online",
    "idle", or "dnd" string per platform.
    If a user is offline or invisible, the corresponding
    field is not present.

    Attributes
    ----------
    desktop: APINullable[:class:`str`]
        The user's status set for an active desktop
        (Windows, Linux, Mac) application session
    mobile: APINullable[:class:`str`]
        The user's status set for an active mobile
        (iOS, Android) application session
    web: APINullable[:class:`str`]
        The user's status set for an active web
        (browser, bot account) application session
    """
    desktop: APINullable[str] = MISSING
    mobile: APINullable[str] = MISSING
    web: APINullable[str] = MISSING


@dataclass
class PresenceUpdateEvent(APIObject):
    """This event is sent when a user's presence or info,
    such as name or avatar, is updated.

    Attributes
    ----------
    user: :class:`~pincer.objects.user.user.User`
        The user presence is being updated for
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the guild
    status: :class:`str`
        Either "idle", "dnd", "online", or "offline"
    activities: List[:class:`~pincer.objects.events.presence.Activity`]
        User's current activities
    client_status: :class:`~pincer.objects.events.presence.ClientStatus`
        User's platform-dependent status
    """
    user: User
    status: str
    activities: List[Activity]
    client_status: ClientStatus
    guild_id: APINullable[Snowflake] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from enum import IntEnum


class Intents(IntEnum):
    """Discord client intents.

    These give your client more permissions.

    .. note::
        The given Intents must also be enabled for your client on
        the discord dashboard.

    Attributes
    ----------
    NONE:
        No intents.
    GUILDS:
        Guilds intent.
    GUILD_MEMBERS:
        Members intent.
    GUILD_BANS:
        Bans intent.
    GUILD_EMOJIS_AND_STICKERS:
        Emoji and Sticker intent.
    GUILD_INTEGRATIONS:
        Integrations intent.
    GUILD_WEBHOOKS:
        Webhooks intent.
    GUILD_INVITES:
        Invites intent.
    GUILD_VOICE_STATES:
        Voice states intent.
    GUILD_PRESENCES:
        Presences intent.
    GUILD_MESSAGES:
        Message intent.
    GUILD_MESSAGE_REACTIONS:
        Reactions to messages intent.
    GUILD_MESSAGE_TYPING:
        Typing to messages intent.
    DIRECT_MESSAGES:
        DM messages intent.
    DIRECT_MESSAGE_REACTIONS:
        DM reaction to messages intent.
    DIRECT_MESSAGE_TYPING:
        DM typing to messages intent.
    """
    NONE = 0
    GUILDS = 1 << 0
    GUILD_MEMBERS = 1 << 1
    GUILD_BANS = 1 << 2
    GUILD_EMOJIS_AND_STICKERS = 1 << 3
    GUILD_INTEGRATIONS = 1 << 4
    GUILD_WEBHOOKS = 1 << 5
    GUILD_INVITES = 1 << 6
    GUILD_VOICE_STATES = 1 << 7
    GUILD_PRESENCES = 1 << 8
    GUILD_MESSAGES = 1 << 9
    GUILD_MESSAGE_REACTIONS = 1 << 10
    GUILD_MESSAGE_TYPING = 1 << 11
    DIRECT_MESSAGES = 1 << 12
    DIRECT_MESSAGE_REACTIONS = 1 << 13
    DIRECT_MESSAGE_TYPING = 1 << 14

    @staticmethod
    def all() -> int:
        """
        :class:`~pincer.objects.app.intents.Intents`:
        Method of all intents
        """
        res = 0

        for intent in list(map(lambda itm: itm.value, Intents)):
            res |= intent

        return res



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from ...utils.types import APINullable
    from ...utils.snowflake import Snowflake
    from ...utils.timestamp import Timestamp


@dataclass
class ThreadMetadata(APIObject):
    """
    Represents a Discord Thread Metadata object

    Attributes
    ----------
    archived: :class:`bool`
        Whether the thread is archived
    auto_archive_duration: :class:`int`
        Duration in minutes to automatically archive the thread
        after recent activity, can be set to: 60, 1440, 4320, 10080
    archive_timestamp: :class:`~pincer.utils.timestamp.Timestamp`
        Timestamp when the thread's archive status was last changed,
        used for calculating recent activity
    locked: :class:`bool`
        Whether the thread is locked; when a thread is locked,
        only users with MANAGE_THREADS can unarchive it
    invitable: APINullable[:class:`bool`]
        Whether non-moderators can add other non-moderators to a thread;
        only available on private threads
    """
    archived: bool
    auto_archive_duration: int
    archive_timestamp: Timestamp
    locked: bool

    invitable: APINullable[bool] = MISSING


@dataclass
class ThreadMember(APIObject):
    """Represents a Discord Thread Member object

    Attributes
    ----------
    join_timestamp: :class:`~pincer.utils.timestamp.Timestamp`
        The time the current user last joined the thread
    flags: :class:`int`
        Any user-thread settings, currently only used for notifications
    id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Id of the thread
    user_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Id of the user
    """
    join_timestamp: Timestamp
    flags: int

    id: APINullable[Snowflake] = MISSING
    user_id: APINullable[Snowflake] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import List, Union, TYPE_CHECKING

from .command_types import AppCommandOptionType, AppCommandType
from ...utils.api_object import APIObject
from ...utils.conversion import convert
from ...utils.snowflake import Snowflake
from ...utils.types import Coro, choice_value_types
from ...utils.types import MISSING

if TYPE_CHECKING:
    from ...utils.types import APINullable
    from ..app.throttle_scope import ThrottleScope


@dataclass
class AppCommandInteractionDataOption(APIObject):
    """Represents a Discord Application Command Interaction Data Option

    Attributes
    ----------
    name: :class:`str`
        The name of the parameter
    value: :class:`str`
        The value of the pair
    type: APINullable[:class:`str`]
        Value of application command option type
    options: APINullable[List[:data:`~pincer.objects.app.command.AppCommandInteractionDataOption`]]
        Present if this option is a group or subcommand
    """
    # noqa: E501
    name: str
    value: APINullable[str] = MISSING
    type: APINullable[AppCommandOptionType] = MISSING
    options: APINullable[
        List[AppCommandInteractionDataOption]] = MISSING


@dataclass
class AppCommandOptionChoice(APIObject):
    """Represents a Discord Application Command Option Choice object

    Attributes
    ----------
    name: :class:`str`
        1-100 character choice name
    value: Union[:data:`~pincer.utils.types.choice_value_types`]
        Value of the choice, up to 100 characters if string
    """
    name: str
    value: choice_value_types


@dataclass
class AppCommandOption(APIObject):
    """Represents a Discord Application Command Option object

    Attributes
    ----------
    type: :class:`~pincer.objects.AppCommandOptionType`
        The type of option
    name: :class:`str`
        1-32 lowercase character name matching `^[\\w-]{1,32}$`
    description: :class:`str`
        1-100 character description
    required: APINullable[:class:`bool`]
        If the parameter is required or optional |default| :data:`False`
    choices: APINullable[List[:class:`~pincer.objects.app.command.AppCommandOptionChoice`]]
        Choices for `STRING`, `INTEGER`, and `NUMBER`
        types for the user to pick from, max 25
    options: APINullable[List[:class:`~pincer.objects.app.command.AppCommandOptionChoice`]]
        If the option is a subcommand or subcommand group type,
        this nested options will be the parameters
    """
    # noqa: E501
    type: AppCommandOptionType
    name: str
    description: str

    required: APINullable[bool] = False
    choices: APINullable[List[AppCommandOptionChoice]] = MISSING
    options: APINullable[List[AppCommandOption]] = MISSING

    def __post_init__(self):
        self.type = AppCommandOptionType(self.type)
        self.choices = convert(
            self.choices,
            AppCommandOptionChoice.from_dict,
            AppCommandOptionChoice
        )
        self.options = convert(
            self.options,
            AppCommandOption.from_dict,
            AppCommandOption
        )


@dataclass
class AppCommand(APIObject):
    """Represents a Discord Application Command object

    Attributes
    ----------
    type: :class:`~pincer.objects.app.command.AppCommandType`
        The type of command, defaults ``1`` if not set
    name: :class:`str`
        1-32 character name
    description: :class:`str`
        1-100 character description for ``CHAT_INPUT`` commands,
        empty string for ``USER`` and ``MESSAGE`` commands
    id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Unique id of the command
    version: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Auto-incrementing version identifier updated during substantial
        record changes
    application_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Unique id of the parent application
    options: APINullable[List[:class:`~pincer.objects.app.command.AppCommandOption`]]
        The parameters for the command, max 25
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Guild id of the command, if not global
    default_permission: APINullable[:class:`bool`]
        Whether the command is enabled by default
        when the app is added to a guild
    """
    # noqa: E501
    type: AppCommandType
    name: str
    description: str

    id: APINullable[Snowflake] = MISSING
    version: APINullable[Snowflake] = MISSING
    application_id: APINullable[Snowflake] = MISSING
    options: APINullable[List[AppCommandOption]] = MISSING
    guild_id: APINullable[Snowflake] = MISSING
    default_permission: APINullable[bool] = True
    default_member_permissions: APINullable[None] = None
    dm_permission: APINullable[None] = None

    _eq_props = [
        "type", "name", "description", "guild_id", "default_permission",
        "options"
    ]

    def __post_init__(self):
        self.id = convert(self.id, Snowflake.from_string)
        self.version = convert(self.version, Snowflake.from_string)
        self.type = AppCommandType(self.type)
        self.application_id = convert(
            self.application_id, Snowflake.from_string
        )

        self.options = convert(
            self.options,
            AppCommandOption.from_dict,
            AppCommandOption
        )
        self.guild_id = convert(self.guild_id, Snowflake.from_string)

        self.options = [] if self.options is MISSING else self.options

    def __eq__(self, other: Union[AppCommand, ClientCommandStructure]):
        if isinstance(other, ClientCommandStructure):
            other = other.app

        return all(
            self.__getattribute__(prop) == other.__getattribute__(prop)
            for prop in self._eq_props
        )

    def __hash__(self):
        return hash((self.id, self.name, self.description, self.guild_id))

    def add_option(self, option: AppCommandOption):
        """Add a new option field to the current application command.

        Parameters
        ----------
        option : :class:`~pincer.objects.app.command.AppCommandOption`
            The option which will be appended.
        """
        if self.options:
            self.options.append(option)
        else:
            self.options = [option]


@dataclass
class ClientCommandStructure:
    """Represents the structure of how the client saves the existing
    commands in the register.

    Attributes
    ----------
    app: :class:`~pincer.objects.app.command.AppCommand`
        The command application.
    call: :class:`~pincer.utils.types.Coro`
        The coroutine which should be called when the command gets
        executed.
    cooldown: :class:`int`
        Amount of times for cooldown
    cooldown_scale: :class:`float`
        Search time for cooldown
    cooldown_scope: :class:`~pincer.objects.app.throttle_scope.ThrottleScope`
        The type of cooldown
    """
    app: AppCommand
    call: Coro
    cooldown: int
    cooldown_scale: float
    cooldown_scope: ThrottleScope



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.


from __future__ import annotations

from json import dumps, loads
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Dict, Optional, Union


class GatewayDispatch:
    """Represents a websocket message.

    Attributes
    ----------
    op: :class:`int`
        The discord opcode which represents what the message means.
    data: Optional[Union[:class:`int`, Dict[:class:`str`, Any]]]
        The event data that has been sent/received.
    seq: Optional[:class:`int`]
        The sequence number of a message, which can be used
        for resuming sessions and heartbeats.
    event_name: Optional[:class:`str`]
        The event name for the payload.
    """

    def __init__(
            self,
            op: int,
            data: Optional[Union[int, Dict[str, Any]]],
            seq: Optional[int] = None,
            name: Optional[str] = None
    ):
        self.op: int = op
        self.data: Optional[Union[int, Dict[str, Any]]] = data
        self.seq: Optional[int] = seq
        self.event_name: Optional[str] = name

    def __str__(self) -> str:
        return dumps(
            dict(
                op=self.op,
                d=self.data,
                s=self.seq,
                t=self.event_name
            )
        )

    @classmethod
    def from_string(cls, payload: str) -> GatewayDispatch:
        """Parses a given payload from a string format
        and returns a GatewayDispatch.

        Parameters
        ----------
        payload : :class:`str`
            The payload to parse.

        Returns
        -------
        :class:`~pincer.core.dispatch.GatewayDispatch`
            The new class.
        """
        payload: Dict[str, Union[int, str, Dict[str, Any]]] = loads(payload)
        return cls(
            payload.get("op"),
            payload.get("d"),
            payload.get("s"),
            payload.get("t")
        )



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils import APIObject

if TYPE_CHECKING:
    from typing import Optional

    from ..guild import Guild
    from ..user.user import User
    from ...utils.snowflake import Snowflake
    from ...utils.timestamp import Timestamp


@dataclass
class GuildTemplate(APIObject):
    """Represents a code that when used,
    creates a guild based on a snapshot of an existing guild.

    Attributes
    ----------
    code: :class:`str`
        the template code (unique ID)
    name: :class:`str`
        template name
    description: Optional[:class:`str`]
        the description for the template
    usage_count: :class:`int`
        number of times this template has been used
    creator_id: :class:`~pincer.utils.snowflake.Snowflake`
        the ID of the user who created the template
    creator: :class:`~pincer.objects.user.user.User`
        user who created the template
    created_at: :class:`~pincer.utils.timestamp.Timestamp`
        when this template was created
    updated_at: :class:`~pincer.utils.timestamp.Timestamp`
        when this template was last synced to the source guild
    source_guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        the ID of the guild this template is based on
    serialized_source_guild: :class:`~pincer.objects.guild.guild.Guild`
        the guild snapshot this template contains
    is_dirty: Optional[:class:`bool`]
        whether the template has unsynced changes
    """
    code: str
    name: str
    description: Optional[str]
    usage_count: int
    creator_id: Snowflake
    creator: User
    created_at: Timestamp
    updated_at: Timestamp
    source_guild_id: Snowflake
    serialized_source_guild: Guild
    is_dirty: Optional[bool]



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a user explicitly removes all reactions from a message."""

from ..core.dispatch import GatewayDispatch
from ..objects.events.message import MessageReactionRemoveAllEvent
from ..utils.conversion import construct_client_dict


async def message_reaction_remove_all_middleware(
    self,
    payload: GatewayDispatch
):
    """|coro|

    Middleware for ``on_message_reaction_remove_all`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the message reaction remove all event.


    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.events.message.MessageReactionRemoveAllEvent`]]
        ``on_message_reaction_remove_all`` and an ``MessageReactionRemoveAllEvent``
    """

    return "on_message_reaction_remove_all", [
        MessageReactionRemoveAllEvent.from_dict(
            construct_client_dict(self, payload.data)
        )
    ]


def export():
    return message_reaction_remove_all_middleware



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Union

from ..objects.app.interaction_flags import InteractionFlags
from ..objects.message.embed import Embed
from ..objects.message.file import File
from ..objects.message.message import Message

PILLOW_IMPORT = True

try:
    from PIL.Image import Image
except (ModuleNotFoundError, ImportError):
    PILLOW_IMPORT = False

if TYPE_CHECKING:
    from pincer import Client

MessageConvertable = Union[Embed, Message, str]


def convert_message(
        client: Client,
        message: MessageConvertable
) -> Message:
    """Converts a message to a Message object"""
    if isinstance(message, Embed):
        message = Message(embeds=[message])
    elif PILLOW_IMPORT and isinstance(message, (File, Image)):
        message = Message(attachments=[message])
    elif not isinstance(message, Message):
        message = Message(message) if message else Message(
            client.received_message,
            flags=InteractionFlags.EPHEMERAL
        )
    return message



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

import logging
from asyncio import sleep
from typing import TYPE_CHECKING

from websockets.exceptions import ConnectionClosedOK

from . import __package__
from ..core.dispatch import GatewayDispatch
from ..exceptions import HeartbeatError

if TYPE_CHECKING:
    from typing import Optional

    from websockets.legacy.client import WebSocketClientProtocol


_log = logging.getLogger(__package__)


class Heartbeat:
    """The heartbeat of the websocket connection.

    This is what lets the server and client know that they are still
    both online and properly connected.
    """
    __heartbeat: float = 0
    __sequence: Optional[int] = None

    @classmethod
    async def __send(cls, socket: WebSocketClientProtocol):
        """|Coro|
        Sends a heartbeat to the API gateway.

        Parameters
        ----------
        socket : :class:`~ws:websockets.legacy.client.WebSocketClientProtocol`
            The socket to send the heartbeat to.
        """
        _log.debug("Sending heartbeat (seq: %s)", str(cls.__sequence))
        try:
            await socket.send(str(GatewayDispatch(1, cls.__sequence)))
        except ConnectionClosedOK:
            _log.error(
                "Sending heartbeat failed. Ignoring failure... "
                "Client should automatically resolve this issue. "
                "If a crash occurs please create an issue on our github! "
                "(https://github.com/Pincer-org/Pincer)"
            )

    @classmethod
    def get(cls) -> float:
        """Get the current heartbeat.

        Returns
        -------
        :class:`float`
            The current heartbeat of the client.
            |default| ``0`` (client has not initialized the heartbeat yet.)
        """
        return cls.__heartbeat

    @classmethod
    async def handle_hello(
            cls,
            socket: WebSocketClientProtocol,
            payload: GatewayDispatch
    ):
        """|coro|

        Handshake between the discord API and the client.
        Retrieve the heartbeat for maintaining a connection.

        Parameters
        ----------
        socket : :class:`~ws:websockets.legacy.client.WebSocketClientProtocol`
            The socket to send the heartbeat to.
        payload : :class:`~pincer.core.dispatch.GatewayDispatch`
            The received hello message from the Discord gateway.

        Raises
        ------
        HeartbeatError
            No ``heartbeat_interval`` is present.
        """
        _log.debug("Handling initial discord hello websocket message.")
        cls.__heartbeat = payload.data.get("heartbeat_interval")

        if not cls.__heartbeat:
            _log.error(
                "No `heartbeat_interval` is present. Has the API changed? "
                "(payload: %s)", payload
            )

            raise HeartbeatError(
                "Discord hello is missing `heartbeat_interval` in payload."
                "Because of this the client can not maintain a connection. "
                "Check logging for more information."
            )

        cls.__heartbeat /= 1000

        _log.debug(
            "Maintaining a connection with heartbeat: %s", cls.__heartbeat
        )

        if Heartbeat.__sequence:
            await socket.send(
                str(GatewayDispatch(6, cls.__sequence, seq=cls.__sequence))
            )

        else:
            await cls.__send(socket)

    @classmethod
    async def handle_heartbeat(cls, socket: WebSocketClientProtocol, _):
        """|coro|

        Handles a heartbeat, which means that it rests
        and then sends a new heartbeat.

        Parameters
        ----------
        socket : :class:`~ws:websockets.legacy.client.WebSocketClientProtocol`
            The socket to send the heartbeat to.
        _ :
            Filling param for auto event handling.
        """
        _log.debug("Resting heart for %is", cls.__heartbeat)
        await sleep(cls.__heartbeat)
        await cls.__send(socket)

    @classmethod
    def update_sequence(cls, seq: int):
        """
        Update the heartbeat sequence.

        Parameters
        ----------
        seq : :class:`int`
            The new heartbeat sequence to be updated with.
        """
        _log.debug("Updating heartbeat sequence...")
        cls.__sequence = seq



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from re import match
from typing import TYPE_CHECKING

from ...exceptions import InvalidUrlError, EmbedFieldError
from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Iterable, Union, Optional

    from ...utils.types import APINullable


def _field_size(_field: str) -> int:
    """
    The Discord API removes white space
        when counting the length of a field.

    :param _field:
        The field.

    :return:
        Length of the string without white space.
    """
    return 0 if _field == MISSING else len(_field.strip())


def _is_valid_url(url: str) -> bool:
    """
    Checks whether the url is a proper and valid url.
    (matches for http and attachment protocol.

    :param url:
        The url which must be checked.

    :return:
        Whether the provided url is valid.
    """
    stmt = (
        r"(http[s]|attachment)"
        r"?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|"
        r"(?:%[0-9a-fA-F][0-9a-fA-F]))+"
    )

    return bool(match(stmt, url))


def _check_if_valid_url(url: str):
    """Checks if the provided url is valid.

    Raises
    ------
    :class:`~pincer.exceptions.InvalidUrlError`
        if the url didn't match the url regex.
        (which means that it was malformed or didn't match the http/attachment
        protocol).
    """
    if not _is_valid_url(url):
        raise InvalidUrlError(
            "Url was malformed or wasn't of protocol http(s)/attachment."
        )


@dataclass
class EmbedAuthor:
    """Representation of the Embed Author class

    Attributes
    ----------
    name: APINullable[:class:`str`]
        Name of the author
    url: APINullable[:class:`str`]
        Url of the author
    icon_url: APINullable[:class:`str`]
        Url of the author icon
    proxy_icon_url: APINullable[:class:`str`]
        A proxied url of the author icon
    """
    icon_url: APINullable[str] = MISSING
    name: APINullable[str] = MISSING
    proxy_icon_url: APINullable[str] = MISSING
    url: APINullable[str] = MISSING

    def __post_init__(self):  # stop documenting special methods
        if _field_size(self.name) > 256:
            raise EmbedFieldError.from_desc("Author name", 256, len(self.name))

        _check_if_valid_url(self.url)


@dataclass
class EmbedImage:
    """Representation of the Embed Image class

    Attributes
    ----------
    url: APINullable[:class:`str`]
        Source url of the image
    proxy_url: APINullable[:class:`str`]
        A proxied url of the image
    height: APINullable[:class:`int`]
        Height of the image
    width: APINullable[:class:`int`]
        Width of the image
    """

    url: APINullable[str] = MISSING
    proxy_url: APINullable[str] = MISSING
    height: APINullable[int] = MISSING
    width: APINullable[int] = MISSING

    def __post_init__(self):
        _check_if_valid_url(self.url)


@dataclass
class EmbedProvider:
    """Representation of the Provider class

    Attributes
    ----------
    name: APINullable[:class:`str`]
        Name of the provider
    url: APINullable[:class:`str`]
        Url of the provider
    """
    name: APINullable[str] = MISSING
    url: APINullable[str] = MISSING


@dataclass
class EmbedThumbnail:
    """Representation of the Embed Thumbnail class

    Attributes
    ----------
    url: APINullable[:class:`str`]
        Source url of the thumbnail
    proxy_url: APINullable[:class:`str`]
        A proxied url of the thumbnail
    height: APINullable[:class:`int`]
        Height of the thumbnail
    width: APINullable[:class:`int`]
        Width of the thumbnail
    """

    url: APINullable[str] = MISSING
    proxy_url: APINullable[str] = MISSING
    height: APINullable[int] = MISSING
    width: APINullable[int] = MISSING

    def __post_init__(self):
        _check_if_valid_url(self.url)


@dataclass
class EmbedVideo:
    """Representation of the Embed Video class

    Attributes
    ----------
    url: APINullable[:class:`str`]
        Source url of the video
    proxy_url: APINullable[:class:`str`]
        A proxied url of the video
    height: APINullable[:class:`int`]
        Height of the video
    width: APINullable[:class:`int`]
        Width of the video
    """
    height: APINullable[int] = MISSING
    url: APINullable[str] = MISSING
    proxy_url: APINullable[str] = MISSING
    width: APINullable[int] = MISSING


@dataclass
class EmbedFooter:
    """Representation of the Embed Footer class

    Attributes
    ----------
    text: :class:`str`
        Footer text
    icon_url: APINullable[:class:`str`]
        Url of the footer icon
    proxy_icon_url: APINullable[:class:`str`]
        A proxied url of the footer icon

    Raises
    ------
    EmbedFieldError:
        Text is longer than 2048 characters
    """

    text: str

    icon_url: APINullable[str] = MISSING
    proxy_icon_url: APINullable[str] = MISSING

    def __post_init__(self):
        if _field_size(self.text) > 2048:
            raise EmbedFieldError.from_desc(
                "Footer text", 2048, len(self.text)
            )


@dataclass
class EmbedField:
    """Representation of the Embed Field class

    Attributes
    ----------
    name: :class:`str`
        The name of the field
    value: :class:`str`
        The text in the field
    inline: APINullable[:class:`bool`]
        Whether or not this field should display inline

    Raises
    ------
    EmbedFieldError:
        Name is longer than 256 characters
    EmbedFieldError:
        Description is longer than 1024 characters
    """

    name: str
    value: str

    inline: APINullable[bool] = MISSING

    def __post_init__(self):
        if _field_size(self.name) > 256:
            raise EmbedFieldError.from_desc(
                "Field name", 256, len(self.name)
            )

        if _field_size(self.value) > 1024:
            raise EmbedFieldError.from_desc(
                "Field value", 1024, len(self.value)
            )


# TODO: Handle Bad Request if embed that is too big is sent
# https://discord.com/developers/docs/resources/channel#embed-limits
# Currently ignored since I don't think it would make sense to put
# This with the Embed class
@dataclass
class Embed(APIObject):
    """Representation of the discord Embed class

    Attributes
    ----------
    title: APINullable[:class:`str`]
        Embed title.
    description: APINullable[:class:`str`]
        Embed description.
    color: APINullable[:class:`int`]
        Embed color code.
    fields: List[:class:`~pincer.objects.message.embed.EmbedField`]
        Fields information.
    footer: APINullable[:class:`~pincer.objects.message.embed.EmbedFooter`]
        Footer information.
    image: APINullable[:class:`~pincer.objects.message.embed.EmbedImage`]
        Image information.
    provider: APINullable[:class:`~pincer.objects.message.embed.EmbedProvider`]
        Provider information.
    thumbnail: APINullable[:class:`~pincer.objects.message.embed.EmbedThumbnail`]
        Thumbnail information.
    timestamp: APINullable[:class:`str`]
        Timestamp of embed content in ISO format.
    url: APINullable[:class:`str`]
        Embed url.
    video: APINullable[:class:`~pincer.objects.message.embed.EmbedVideo`]
        Video information.
    type: APINullable[:class:`int`]
        type of message
    """
    # noqa: E501

    title: APINullable[str] = MISSING
    description: APINullable[str] = MISSING
    color: APINullable[int] = MISSING
    fields: list[EmbedField] = field(default_factory=list)
    footer: APINullable[EmbedFooter] = MISSING
    image: APINullable[EmbedImage] = MISSING
    provider: APINullable[EmbedProvider] = MISSING
    thumbnail: APINullable[EmbedThumbnail] = MISSING
    timestamp: APINullable[str] = MISSING
    author: APINullable[EmbedAuthor] = MISSING
    url: APINullable[str] = MISSING
    video: APINullable[EmbedVideo] = MISSING
    type: APINullable[int] = MISSING

    def __post_init__(self):
        if _field_size(self.title) > 256:
            raise EmbedFieldError.from_desc(
                "Embed title", 256, len(self.title)
            )

        if _field_size(self.description) > 4096:
            raise EmbedFieldError.from_desc(
                "Embed description", 4096, len(self.description)
            )

        if len(self.fields) > 25:
            raise EmbedFieldError.from_desc(
                "Embed field", 25, len(self.fields)
            )

    def set_timestamp(self, time: datetime) -> Embed:
        """Discord uses iso format for time stamps.
        This function will set the time to that format.

        Parameters
        ----------
        time : :class:`datetime.datetime`
            The datetime to set the timestamp to.

        Returns
        -------
        :class:`~pincer.objects.message.embed.Embed`
            The new embed object.
        """
        self.timestamp = time.isoformat()

        return self

    def set_author(
            self,
            icon_url: APINullable[str] = MISSING,
            name: APINullable[str] = MISSING,
            proxy_icon_url: APINullable[str] = MISSING,
            url: APINullable[str] = MISSING
    ) -> Embed:
        """Set the author message for the embed. This is the top
        field of the embed.

        Parameters
        ----------
        icon_url: APINullable[:class:`str`]
            The icon which will be next to the author name.
        name: APINullable[:class:`str`]
            The name for the author (so the message).
        proxy_icon_url: APINullable[:class:`str`]
            A proxied url of the author icon.
        url: APINullable[:class:`str`]
            The url for the author name, this will make the
            name field a link/url.

        Returns
        -------
        :class:`~pincer.objects.message.embed.Embed`
            The new embed object.
        """

        self.author = EmbedAuthor(
            icon_url=icon_url,
            name=name,
            proxy_icon_url=proxy_icon_url,
            url=url
        )

        return self

    def set_image(
            self,
            url: APINullable[str] = MISSING,
            proxy_url: APINullable[str] = MISSING,
            height: APINullable[int] = MISSING,
            width: APINullable[int] = MISSING
    ) -> Embed:
        """Sets an image for your embed.

        Parameters
        ----------
        url: APINullable[:class:`str`]
            Source url of the video
        proxy_url: APINullable[:class:`str`]
            A proxied url of the video
        height: APINullable[:class:`int`]
            Height of the video
        width: APINullable[:class:`int`]
            Width of the video

        Returns
        -------
        :class:`~pincer.objects.message.embed.Embed`
            The new embed object.
        """
        self.image = EmbedImage(
            height=height,
            url=url,
            proxy_url=proxy_url,
            width=width
        )

        return self

    def set_thumbnail(
            self,
            height: APINullable[int] = MISSING,
            url: APINullable[str] = MISSING,
            proxy_url: APINullable[str] = MISSING,
            width: APINullable[int] = MISSING
    ) -> Embed:  # ? its normally smaller in the corner?
        """Sets the thumbnail of the embed.
        This image is bigger than the ``image`` property.

        url: APINullable[:class:`str`]
            Source url of the video
        proxy_url: APINullable[:class:`str`]
            A proxied url of the video
        height: APINullable[:class:`int`]
            Height of the video
        width: APINullable[:class:`int`]
            Width of the video

        Returns
        -------
        :class:`~pincer.objects.message.embed.Embed`
            The new embed object.
        """
        self.thumbnail = EmbedThumbnail(
            height=height,
            url=url,
            proxy_url=proxy_url,
            width=width
        )

        return self

    def set_footer(
            self,
            text: str,
            icon_url: APINullable[str] = MISSING,
            proxy_icon_url: APINullable[str] = MISSING
    ) -> Embed:
        """
        Sets the embed footer. This is at the bottom of your embed.

        Parameters
        ----------
        text: :class:`str`
            Footer text
        icon_url: APINullable[:class:`str`]
            Url of the footer icon
        proxy_icon_url: APINullable[:class:`str`]
            A proxied url of the footer icon

        Returns
        -------
        :class:`~pincer.objects.message.embed.Embed`
            The new embed object.
        """
        self.footer = EmbedFooter(
            text=text,
            icon_url=icon_url,
            proxy_icon_url=proxy_icon_url
        )

        return self

    def add_field(
            self,
            name: str,
            value: str,
            inline: APINullable[bool] = MISSING
    ) -> Embed:
        """Adds a field to the embed.
        An embed can contain up to 25 fields.

        Parameters
        ----------
        name: :class:`str`
            The name of the field
        value: :class:`str`
            The text in the field
        inline: APINullable[:class:`bool`]
            Whether or not this field should display inline

        Raises
        ------
        EmbedFieldError:
            Raised when there are more than 25 fields in the embed
        """
        _field = EmbedField(
            name=name,
            value=value,
            inline=inline
        )

        if len(self.fields) > 25:
            raise EmbedFieldError.from_desc(
                "Embed field", 25, len(self.fields) + 1
            )

        self.fields += [_field]

        return self

    def add_fields(
            self,
            field_list: Union[Dict[Any, Any], Iterable[Iterable[Any, Any]]],
            checks: Optional[Callable[[Any], Any]] = bool,
            map_title: Optional[Callable[[Any], str]] = str,
            map_values: Optional[Callable[[Any], str]] = str,
            inline: bool = True
    ) -> Embed:
        """Add multiple fields from a list,
        dict or generator of fields with possible mapping.

        Parameters
        ----------
        field_list: Union[Dict[Any, Any], Iterable[Iterable[Any, Any]]]
            A iterable or generator of the fields to add.
            If the field_list type is a dictionary, will take items.
        checks: Optional[Callable[[Any], Any]]
            A filter function to remove embed fields.
        map_title: Optional[Callable[[Any], :class:`str`]]
            A transform function to change the titles.
        map_values: Optional[Callable[[Any], :class:`str`]]
            A transform function to change the values.
        inline: :class:`bool`
            Whether to create grid or each field on a new line.

        Raises
        ------
        EmbedFieldError:
            Raised when there are more than 25 fields in the embed

        Returns
        -------
        :class:`~pincer.objects.message.embed.Embed`
            The new embed object.
        """

        if isinstance(field_list, dict):
            field_list: Iterable[Iterable[Any, Any]] = field_list.items()

        for field_name, field_value in field_list:
            val = (
                map_values(field_value)
                if not isinstance(field_value, tuple)
                else map_values(*field_value)
            )

            if checks(val):
                self.add_field(
                    name=map_title(field_name),
                    value=val,
                    inline=inline
                )

        return self



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from .throttle_scope import ThrottleScope
from ...exceptions import CommandCooldownError
from ...utils.slidingwindow import SlidingWindow

if TYPE_CHECKING:
    from typing import Dict, Optional

    from ...utils.types import Coro
    from ..message.context import MessageContext


class ThrottleInterface(ABC):
    """An ABC for throttling."""
    throttle: Dict[Coro, Dict[Optional[str], SlidingWindow]] = {}

    @staticmethod
    @abstractmethod
    def handle(ctx: MessageContext, **kwargs):
        raise NotImplementedError


class DefaultThrottleHandler(ThrottleInterface, ABC):
    """The default throttlehandler based off the
    :class:`~pincer.objects.app.throttling.ThrottleInterface` ABC
    """
    __throttle_scopes = {
        ThrottleScope.GLOBAL: None,
        ThrottleScope.GUILD: "guild_id",
        ThrottleScope.CHANNEL: "channel_id",
        ThrottleScope.USER: "author.user.id"
    }

    @staticmethod
    def get_key_from_scope(ctx: MessageContext) -> Optional[int]:
        """Retrieve the the appropriate key from the context through the
        throttle scope.

        Parameters
        ----------
        ctx : :class:`~pincer.objects.message.context.MessageContext`
            The context to retrieve with

        Returns
        -------
        Optional[:class:`int`]
            The throttlescope enum
        """
        scope = DefaultThrottleHandler.__throttle_scopes[
            ctx.command.cooldown_scope]

        if not scope:
            return None

        last_obj = ctx

        for attr in scope.split("."):
            last_obj = getattr(last_obj, attr)

        return last_obj

    @staticmethod
    def init_throttler(ctx: MessageContext, throttle_key: Optional[int]):
        DefaultThrottleHandler.throttle[ctx.command.call][throttle_key] \
            = SlidingWindow(ctx.command.cooldown, ctx.command.cooldown_scale)

    @staticmethod
    def handle(ctx: MessageContext, **kwargs):
        if ctx.command.cooldown <= 0:
            return

        throttle_key = DefaultThrottleHandler.get_key_from_scope(ctx)
        group = DefaultThrottleHandler.throttle.get(ctx.command.call)
        window_slider = group.get(throttle_key) if group is not None else None

        if window_slider:
            if not window_slider.allow():
                raise CommandCooldownError(
                    f"Cooldown for command {ctx.command.app.name} not met!",
                    ctx
                )
        else:
            DefaultThrottleHandler.init_throttler(ctx, throttle_key)
            DefaultThrottleHandler.handle(ctx)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
Sent when a new user joins a guild. The inner payload is a guild member object
with an extra ``guild_id`` key.
"""

from ..core.dispatch import GatewayDispatch
from ..objects.events.guild import GuildMemberAddEvent
from ..utils import Coro
from ..utils.conversion import construct_client_dict


async def guild_member_add_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_guild_member_add`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the guild member add event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.events.guild.GuildMemberAddEvent`]]
        ``on_guild_member_add`` and a ``GuildMemberAddEvent``
    """

    return "on_guild_member_add", [
        GuildMemberAddEvent.from_dict(construct_client_dict(self, payload.data))
    ]


def export() -> Coro:
    return guild_member_add_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import Any, List, Optional

    from ...utils.snowflake import Snowflake
    from ...utils.types import APINullable
    from ..guild.channel import Channel
    from ..guild.webhook import Webhook
    from ..user.integration import Integration
    from ..user.user import User


class AuditLogEvent(IntEnum):
    """Audit log action type.
    This represents the action which got performed,
    and logged.

    Attributes
    ----------
    GUILD_UPDATE:
        Guild updated.
    CHANNEL_CREATE:
        Channel created.
    CHANNEL_UPDATE:
        Channel updated.
    CHANNEL_DELETE:
        Channel deleted.
    CHANNEL_OVERWRITE_CREATE:
        Channel overwrites created.
    CHANNEL_OVERWRITE_UPDATE:
        Channel overwrites updated.
    CHANNEL_OVERWRITE_DELETE:
        Channel overwrites deleted.
    MEMBER_KICK:
        Member kicked.
    MEMBER_PRUNE:
        Members pruned.
    MEMBER_BAN_ADD:
        Member banned.
    MEMBER_BAN_REMOVE:
        Member unbanned.
    MEMBER_UPDATE:
        Member edit.
    MEMBER_ROLE_UPDATE:
        Member role change.
    MEMBER_MOVE:
        Member voice channel move.
    MEMBER_DISCONNECT:
        Member voice channel disconnect.
    BOT_ADD:
        Bot added.
    ROLE_CREATE:
        Role created.
    ROLE_UPDATE:
        Role updated.
    ROLE_DELETE:
        Role deleted.
    INVITE_CREATE:
        Invite created.
    INVITE_UPDATE:
        Invite updated.
    INVITE_DELETE:
        Invite deleted.
    WEBHOOK_CREATE:
        Webhook created.
    WEBHOOK_UPDATE:
        Webhook updated.
    WEBHOOK_DELETE:
        Webhook deleted.
    EMOJI_CREATE:
        Emoji created.
    EMOJI_UPDATE:
        Emoji updated.
    EMOJI_DELETE:
        Emoji deleted.
    MESSAGE_DELETE:
        Message deleted.
    MESSAGE_BULK_DELETE:
        Message bulk delete.
    MESSAGE_PIN:
        Message pinned.
    MESSAGE_UNPIN:
        Message unpinned.
    INTEGRATION_CREATE:
        Integration created.
    INTEGRATION_UPDATE:
        Integration updated.
    INTEGRATION_DELETE:
        Integration deleted.
    STAGE_INSTANCE_CREATE:
        Stage instance created.
    STAGE_INSTANCE_UPDATE:
        Stage instance updated.
    STAGE_INSTANCE_DELETE:
        Stage instance deleted.
    STICKER_CREATE:
        Sticker created.
    STICKER_UPDATE:
        Sticker updated.
    STICKER_DELETE:
        Sticker deleted.
    THREAD_CREATE1:
        Thread created.
    THREAD_UPDATE1:
        Thread updated.
    THREAD_DELETE1:
        Thread deleted.
    """
    GUILD_UPDATE = 1
    CHANNEL_CREATE = 10
    CHANNEL_UPDATE = 11
    CHANNEL_DELETE = 12
    CHANNEL_OVERWRITE_CREATE = 13
    CHANNEL_OVERWRITE_UPDATE = 14
    CHANNEL_OVERWRITE_DELETE = 15
    MEMBER_KICK = 20
    MEMBER_PRUNE = 21
    MEMBER_BAN_ADD = 22
    MEMBER_BAN_REMOVE = 23
    MEMBER_UPDATE = 24
    MEMBER_ROLE_UPDATE = 25
    MEMBER_MOVE = 26
    MEMBER_DISCONNECT = 27
    BOT_ADD = 28
    ROLE_CREATE = 30
    ROLE_UPDATE = 31
    ROLE_DELETE = 32
    INVITE_CREATE = 40
    INVITE_UPDATE = 41
    INVITE_DELETE = 42
    WEBHOOK_CREATE = 50
    WEBHOOK_UPDATE = 51
    WEBHOOK_DELETE = 52
    EMOJI_CREATE = 60
    EMOJI_UPDATE = 61
    EMOJI_DELETE = 62
    MESSAGE_DELETE = 72
    MESSAGE_BULK_DELETE = 73
    MESSAGE_PIN = 74
    MESSAGE_UNPIN = 75
    INTEGRATION_CREATE = 80
    INTEGRATION_UPDATE = 81
    INTEGRATION_DELETE = 82
    STAGE_INSTANCE_CREATE = 83
    STAGE_INSTANCE_UPDATE = 84
    STAGE_INSTANCE_DELETE = 85
    STICKER_CREATE = 90
    STICKER_UPDATE = 91
    STICKER_DELETE = 92
    THREAD_CREATE = 110
    THREAD_UPDATE = 111
    THREAD_DELETE = 112


@dataclass
class AuditLogChange(APIObject):
    """Representation of Discord Audit Log Change object

    Attributes
    ----------
    new_value: Any
        New value of the key
    old_value: Any
        Old value of the key
    key: :class:`str`
        Name of audit log change key
    """
    new_value: Any
    old_value: Any
    key: str


@dataclass
class AuditEntryInfo(APIObject):
    """Represents Discord Optional Audit Entry Info

    Attributes
    ----------
    delete_member_days: :class:`str`
        Number of days after which inactive members were kicked
    members_removed: :class:`str`
        Number of members removed by the prune
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        Channel in which the entities were targeted
    message_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the message that was targeted
    count: :class:`str`
        Number of entities that were targeted
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the overwritten entity
    type: :class:`str`
        Type of overwritten entity - "0" for "role" or "1" for "member"
    role_name: :class:`str`
        Name of the role if type is "0" (not present if type is "1")
    """
    delete_member_days: str
    members_removed: str
    channel_id: Snowflake
    message_id: Snowflake
    count: str
    id: Snowflake
    type: str
    role_name: str


@dataclass
class AuditLogEntry(APIObject):
    """Represents a Discord Audit Log Entry object.

    Attributes
    ----------
    target_id: Optional[:class:`str`]
        Id of the affected entity x(webhook, user, role, etc.)
    user_id: :class:`~pincer.utils.snowflake.Snowflake`
        The user who made the changes
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the entry
    action_type: :class:`~pincer.objects.guild.audit_log.AuditLogEvent`
        Type of action that occurred
    changes: APINullable[List[:class:`~pincer.objects.guild.audit_log.AuditLogChange`]]
        Changes made to the target_id
    options: APINullable[List[:class:`~pincer.objects.guild.audit_log.AuditLogChange`]]
        Additional info for certain action types
    reason: APINullable[:class:`str`]
        The reason for the change x(0-512 characters)
    """
    # noqa: E501
    target_id: Optional[str]
    user_id: Optional[Snowflake]
    id: Snowflake
    action_type: AuditLogEvent

    changes: APINullable[List[AuditLogChange]] = MISSING
    options: APINullable[AuditEntryInfo] = MISSING
    reason: APINullable[str] = MISSING


@dataclass
class AuditLog(APIObject):
    """Represents a Discord Audit Log object.

    webhooks: List[:class:`~pincer.objects.guild.webhook.Webhook`]
        list of webhooks found in the audit log
    users: List[:class:`~pincer.objects.user.user.User`]
        list of users found in the audit log
    audit_log_entries: List[:class:`~pincer.objects.guild.audit_log.AuditLogEntry`]
        list of audit log entries
    integrations: List[:class:`~pincer.objects.user.integration.Integration`]
        list of partial integration objects
    threads: List[:class:`~pincer.objects.guild.channel.Channel`]
        list of threads found in the audit log
    """
    # noqa: E501
    webhooks: List[Webhook]
    users: List[User]
    audit_log_entries: List[AuditLogEntry]
    integrations: List[Integration]
    threads: List[Channel]



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import Optional

    from ...utils.types import APINullable
    from ...utils.snowflake import Snowflake


@dataclass
class Attachment(APIObject):
    """Represents a Discord Attachment object

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Attachment id
    filename: :class:`str`
        Name of file attached
    content_type: :class:`int`
        The attachment's data type
    size: :class:`str`
        Size of file in bytes
    url: :class:`str`
        Source url of file
    proxy_url: APINullable[:class:`str`]
        A proxied url of file
    height: APINullable[Optional[:class:`int`]]
        Height of file (if image)
    width: APINullable[Optional[:class:`int`]]
        Width of file (if image)
    """
    id: Snowflake
    filename: str
    size: int
    url: str
    proxy_url: str

    content_type: APINullable[str] = MISSING
    height: APINullable[Optional[int]] = MISSING
    width: APINullable[Optional[int]] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""Sent when a stage instance is deleted or closed."""
from ..core.dispatch import GatewayDispatch
from ..objects import StageInstance
from ..utils.conversion import construct_client_dict
from ..utils.types import Coro


def stage_instance_delete_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_stage_instance_delete`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the stage instance delete event

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.guild.stage.StageInstance`]]
        ``on_stage_instance_delete`` and a ``StageInstance``
    """

    return "on_stage_instance_delete", [
        StageInstance.from_dict(construct_client_dict(self, payload.data))
    ]


def export() -> Coro:
    return stage_instance_delete_middleware
